#ifndef _Exp_h_
#define _Exp_h_

#include <string>
#include <iostream>
#include <vector>

using namespace std;

class ExpExc
{
  public:
    ExpExc(const string &s) : message(s) {}
    const string message;
};
        
//
// Exp
// |
// +--LeafExp
// |  |
// |  +--ConstExp
// |  |
// |  +--TimeExp
// |
// +--UnaryExp
// |  |
// |  +--UnaryMinusExp
// |  |
// |  +--SinExp
// |  |
// |  +--CosExp
// |
// +--BinaryExp
//    |
//    +--PlusExp
//    |
//    +--TimesExp
//    |
//    +--GreaterExp
//    |
//    +--LessExp
// 

class Exp
{
  public:

    virtual ~Exp();
    virtual int   getNumInputs() const = 0;
    virtual float evaluate(float t) const = 0;
    virtual float devaluate(float t) const = 0;
    virtual void  mutate(float prob);
    virtual void  unparse(ostream &out) const = 0;
    virtual Exp * duplicate() const = 0;

    static Exp *  getRandomNode(int max_depth);
    static Exp *  getRandomNodeZero(int max_depth);

    typedef vector<Exp *> container;

  protected:

    Exp();

    static bool   mutationOccured(float prob);
};

class LeafExp : public Exp
{
  public:

    LeafExp();
    virtual ~LeafExp();
    virtual int   getNumInputs() const;
    virtual void  mutate(float prob);
};

class ConstExp : public LeafExp
{
  public:

    ConstExp(float val);
    virtual ~ConstExp();
    virtual float evaluate(float t) const;
    virtual float devaluate(float t) const;
    virtual void  mutate(float prob);
    virtual void  unparse(ostream &out) const;
    virtual ConstExp * duplicate() const;

  protected:

    float _val;
};

class TimeExp : public LeafExp
{
  public:

    TimeExp();
    virtual ~TimeExp();
    virtual float evaluate(float t) const;
    virtual float devaluate(float t) const;
    virtual void  mutate(float prob);
    virtual void  unparse(ostream &out) const;
    virtual TimeExp * duplicate() const;
};

class UnaryExp : public Exp
{
  public:

    virtual ~UnaryExp();
    virtual int   getNumInputs() const;
    virtual void  mutate(float prob);

  protected:

    UnaryExp(Exp *child);

    Exp *_child;
};

class UnaryMinusExp : public UnaryExp
{
  public:

    UnaryMinusExp(Exp *child);
    virtual ~UnaryMinusExp();
    virtual float evaluate(float t) const;
    virtual float devaluate(float t) const;
    virtual void  mutate(float prob);
    virtual void  unparse(ostream &out) const;
    virtual UnaryMinusExp * duplicate() const;
};

class SinExp : public UnaryExp
{
  public:

    SinExp(Exp *child);
    virtual ~SinExp();
    virtual float evaluate(float t) const;
    virtual float devaluate(float t) const;
    virtual void  mutate(float prob);
    virtual void  unparse(ostream &out) const;
    virtual SinExp * duplicate() const;
};

class CosExp : public UnaryExp
{
  public:

    CosExp(Exp *child);
    virtual ~CosExp();
    virtual float evaluate(float t) const;
    virtual float devaluate(float t) const;
    virtual void  mutate(float prob);
    virtual void  unparse(ostream &out) const;
    virtual CosExp * duplicate() const;
};

class BinaryExp : public Exp
{
  public:

    virtual ~BinaryExp();
    virtual int   getNumInputs() const;
    virtual void  mutate(float prob);

  protected:

    BinaryExp(Exp *lchild, Exp *rchild);

    Exp *_lchild;
    Exp *_rchild;
};

class PlusExp : public BinaryExp
{
  public:

    PlusExp(Exp *lchild, Exp *rchild);
    virtual ~PlusExp();
    virtual float evaluate(float t) const;
    virtual float devaluate(float t) const;
    virtual void  mutate(float prob);
    virtual void  unparse(ostream &out) const;
    virtual PlusExp * duplicate() const;
};

class TimesExp : public BinaryExp
{
  public:

    TimesExp(Exp *lchild, Exp *rchild);
    virtual ~TimesExp();
    virtual float evaluate(float t) const;
    virtual float devaluate(float t) const;
    virtual void  mutate(float prob);
    virtual void  unparse(ostream &out) const;
    virtual TimesExp * duplicate() const;
};

class GreaterExp : public BinaryExp
{
  public:

    GreaterExp(ConstExp *val, Exp *lchild, Exp *rchild);
    virtual ~GreaterExp();
    virtual float evaluate(float t) const;
    virtual float devaluate(float t) const;
    virtual void  mutate(float prob);
    virtual void  unparse(ostream &out) const;
    virtual GreaterExp * duplicate() const;

  protected:

    ConstExp *_val;
};

class LessExp : public BinaryExp
{
  public:

    LessExp(ConstExp *val, Exp *lchild, Exp *rchild);
    virtual ~LessExp();
    virtual float evaluate(float t) const;
    virtual float devaluate(float t) const;
    virtual void  mutate(float prob);
    virtual void  unparse(ostream &out) const;
    virtual LessExp * duplicate() const;

  protected:

    ConstExp *_val;
};

#endif // _Exp_h_

