#include "Genetic.h"

#include <cstdlib>
#include <algorithm>
#include <fstream>

#ifndef _Exp_h_
# include "Exp.h"
#endif

#ifndef _Parser_h_
# include "Parser.h"
#endif

GenePool::GenePool()
{
    _reproductionRate = 0.1f;
    _mutationRate = 0.01;
    _maxPopulation = 500;
}

GenePool::~GenePool()
{
    deleteAll();
}

void
GenePool::write(const string &filename) const
{
    ofstream out(filename.c_str());
    unparse(out);
    out.close();
}

void
GenePool::unparse(ostream &out) const
{
    _sim.unparse(out);

    for (Genome::container::const_iterator i = _genomes.begin(); 
         i != _genomes.end();
         ++i)
    {
        for (Exp::container::const_iterator j = (*i)->exps.begin();
             j != (*i)->exps.end();
             ++j)
        {
            (*j)->unparse(out);
            out << endl;
        }
    }
}

void
GenePool::setMaxPopulation(int max)
{
    _maxPopulation = max;
}

void
GenePool::deleteAll()
{
    //
    // We are responsible for sim's exps
    //
    _sim.exps.clear();
    _sim.deleteAll();

    //
    // _activeJoints is now bogus
    //
    _activeJoints.clear();

    //
    // Free the Genomes
    //
    for (Genome::container::iterator i = _genomes.begin();
         i != _genomes.end();
         ++i)
    {
        delete (*i);
    }
}


int
GenePool::parse(const string &filename)
{
    deleteAll();

    Exp::container exps;
    
    int error = parseFile(filename, _sim.bodies, _sim.joints, exps);
    if (error)
    {
        cerr << "<GenePool::parse> Error parsing file " << filename << endl;
        deleteAll();
        return error;
    }

    for (Joint::container::iterator i = _sim.joints.begin(); 
         i != _sim.joints.end(); 
         ++i)
    {
        if (dynamic_cast<ConstantJoint *>(*i) != 0) continue;
        _activeJoints.push_back(*i);
    }

    if (_activeJoints.size() == 0)
    {
        cerr << "Input error: no active joints to simulate" << endl;
        return -1;
    }

    if (exps.size() % _activeJoints.size() != 0)
    {
        cerr << "Parse error: " << _activeJoints.size() << " active joints, " 
             << exps.size() << " expressions." << endl;
        return -1;
    }

    int population = exps.size() / _activeJoints.size();
    Exp::container::iterator eiter = exps.begin();

    for (int i = 0; i < population; ++i)
    {
        Genome *genome = new Genome();
        for (unsigned int j = 0; j < _activeJoints.size(); j++)
        {
            genome->exps.push_back((*eiter));
            ++eiter;
        }
        _genomes.push_back(genome);
    }

    // this isn't nessecary - it just goes away
    exps.clear();

    return 0;
}

Genome *
Genome::knockBoots(Genome *mom, Genome *dad)
{
    assert(mom && dad);
    assert(mom->exps.size() == dad->exps.size());

    Genome *kid = new Genome();

    for (unsigned int i = 0; i < dad->exps.size(); ++i)
    {
        if (rand() % 2 == 1)
        {
            kid->exps.push_back(dad->exps[i]->duplicate());
        }
        else
        {
            kid->exps.push_back(mom->exps[i]->duplicate());
        }
    }

    //
    // Maybe do some crossing over here?
    //

    return kid;
}

Genome *
GenePool::findParent()
{
    int index  = rand() % _genomes.size();

    int nindex = rand() % _genomes.size();
    index = index < nindex ? index : nindex;

    nindex = rand() % _genomes.size();
    index = index < nindex ? index : nindex;

    nindex = rand() % _genomes.size();
    index = index < nindex ? index : nindex;

    return _genomes[index];
}

void
GenePool::reproduce(int number)
{
    for (int i = 0; i < number; i++)
    {
        Genome *mom = findParent();
        Genome *dad = findParent();
        Genome *kid = Genome::knockBoots(mom, dad);
        _genomes.push_back(kid);
    }
}

void
GenePool::addRandomGenome(int max_depth)
{
    Genome *kid = new Genome();

    for (unsigned int i = 0; i < _activeJoints.size(); i++)
    {
        kid->exps.push_back(Exp::getRandomNodeZero(max_depth));
    }

    _genomes.push_back(kid);
}

void
GenePool::doMutation(float rate)
{
    Genome::container::iterator i = _genomes.begin();
    
    ++i; // leave the best one untouched

    for(; i != _genomes.end(); ++i)
    {
        for (Exp::container::iterator j = (*i)->exps.begin();
             j != (*i)->exps.end();
             ++j)
        {
            (*j)->mutate(rate);
        }
    }
}

void
GenePool::modifyPopulation()
{
    //
    // Create a quarter more genomes
    //
    reproduce((int) (getReproductionRate() * getPopulation()));

    //
    // Add genomes until we have 
    //
    while (getPopulation() < getMaxPopulation())
    {
        addRandomGenome(8);
    }

    //
    // Mututate the population
    //
    doMutation(getMutationRate());
}

void
GenePool::simulateAll()
{
    for (Genome::container::iterator i = _genomes.begin();
         i != _genomes.end();
         ++i)
    {
        (*i)->fitness = 0;

        for (unsigned int j = 0; j < _activeJoints.size(); ++j)
        {
            _activeJoints[j]->target_exp = (*i)->exps[j];
        }

        _sim.reset();
        while (_sim.getCurrentTime() < getSimulationLength())
        {
            _sim.doTimestep();
        }

        // fitness is high jumps...
        (*i)->fitness = 5 * (_sim.getMaximumHeight() - _sim.getGroundLevel());
        // ...distance travelled in x...
        //(*i)->fitness += _sim.bodies[0]->pos.x - (_sim.bodies[0]->_end1orig.x
        //                                + _sim.bodies[0]->_end2orig.x) * 0.5f;

        float max_dtarget = 0.0f;
        for (Joint::container::iterator j = _sim.joints.begin();
             j != _sim.joints.end();
             ++j)
        {
            if ((*j)->max_dtarget > max_dtarget) max_dtarget = (*j)->max_dtarget;
        }

        // ...and small joint target derivatives
        (*i)->fitness -= max_dtarget;

        if (_sim.hadSelfCollision())
        {
            (*i)->fitness -= 100.0f;
        }

        cout << "." << flush;
    }
}    

inline bool fit_cmp(Genome *g1, Genome *g2)
{
    return g1->fitness > g2->fitness;
}

void
GenePool::sortByFitness()
{
    sort(_genomes.begin(), _genomes.end(), fit_cmp);
}

void
GenePool::killTheWeak()
{
    while (_genomes.size() > (unsigned int) (_maxPopulation * 0.75))
    {
        delete _genomes.back();
        _genomes.pop_back();
    }
}   

void
GenePool::doGeneration()
{
    modifyPopulation();
    simulateAll();
    sortByFitness();
    killTheWeak();
}
