#include "Sim.h"

#include <GL/glut.h>
#include <GL/gl.h>
#include <sys/time.h>
#include <unistd.h>

#include <cmath>
#include <vector>

#ifndef _Math_h_
# include "Math.h"
#endif

#ifndef _Body_h_
# include "Body.h"
#endif

#ifndef _Joint_h_
# include "Joint.h"
#endif

#ifndef _Exp_h_
# include "Exp.h"
#endif

#ifndef _Mouse_h_
# include "Mouse.h"
#endif

#ifndef _Parser_h_
# include "Parser.h"
#endif

const float Sim::timestep        = 0.0001f;
const float Sim::mouse_k         = 10.0f;
const float Sim::mouse_b         = 1.0f;
const float Sim::joint_k         = 10000.0f;
const float Sim::joint_b         = 100.0f;
const float Sim::joint_ang_k     = 100.0f;
const float Sim::joint_ang_b     = 1.0f;
const float Sim::gravity         = -9.8f;
const float Sim::ground_k        = 100000.0f;
const float Sim::ground_b        = 0.0f; 
const float Sim::ground_level    = -0.0f;
const float Sim::ground_friction = 6.0f;

Sim::Sim()
{
    reset();
}

Sim::~Sim()
{
    deleteAll();
}

int
Sim::parse(const string &filename)
{
    deleteAll();
    
    int error = parseFile(filename, bodies, joints, exps);
    if (error)
    {
        cerr << "<Sim::parse> Error parsing file " << filename << endl;
        deleteAll();
        return error;
    }

    return 0;
}

void
Sim::deleteAll()
{
    for (Body::container::iterator i = bodies.begin(); i != bodies.end(); ++i)
    {
        delete *i;
    }
    bodies.clear();

    for (Joint::container::iterator i = joints.begin(); i != joints.end(); ++i)
    {
        delete *i;
    }
    joints.clear();

    for (Exp::container::iterator i = exps.begin(); i != exps.end(); ++i)
    {
        delete *i;
    }
    exps.clear();
}

void
Sim::unparse(ostream &out) const
{
    out << "bodies" << endl;
    for (Body::container::const_iterator i = bodies.begin(); 
         i != bodies.end(); 
         ++i)
    {
        (*i)->unparse(out);
        out << endl;
    }

    out << "joints" << endl;
    for (Joint::container::const_iterator i = joints.begin(); 
         i != joints.end(); 
         ++i)
    {
        (*i)->unparse(out);
        out << endl;
    }

    out << "exps" << endl;
    for (Exp::container::const_iterator i = exps.begin(); 
         i != exps.end(); 
         ++i)
    {
        (*i)->unparse(out);
        out << endl;
    }
}

void
Sim::reset()
{
    _currentTime = 0.0f;
    _maxHeight   = ground_level;
    _selfCollided = false;

    for (Body::container::iterator i = bodies.begin(); 
         i != bodies.end(); 
         ++i)
    {
        (*i)->reset();
    }

    for (Joint::container::iterator i = joints.begin(); 
         i != joints.end(); 
         ++i)
    {
        (*i)->reset();
    }
}

float
Sim::getGroundLevel() const
{
    return ground_level;
}

void
Sim::calcConstantForces(float dt)
{
    int j = 0;

    for (Body::container::iterator i = bodies.begin(); i != bodies.end(); ++i)
    {
        Body *b = *i;

        b->F.x = 0;
        b->F.y = gravity * b->M;
        b->N  = 0;

        if (j == 0 && mouseDown)
        {
            float val = 0.001f;
            b->pos = mousePos * val + b->pos * (1.0f - val);
            b->vel = (mousePos - b->pos) * (1.0f - val);
        }

        j += 1;
    }
}

void
Sim::predictFuturePositions(float dt)
{
    static bool xfirst = true;

    for (Body::container::iterator i = bodies.begin(); i != bodies.end(); ++i)
    {
        Body *b = *i;

        //
        // Test integrate position velocity, orientation and angular velocity
        //
        b->fpos      = b->pos + b->vel * dt;
        b->fvel      = b->vel + b->F * (dt * b->oneOverM);

        float dtheta = b->omega * dt;
        b->ftheta    = b->theta + dtheta;
        b->fomega    = b->omega + b->N * dt * b->oneOverI;

        //
        // Compute normalized direction (and normal)
        //
        //b->fdir      = V2f((float) cos(b->ftheta), (float) sin(b->ftheta));
        // sinusoidal circle drawing courtesy of Akin and Simar, 
        // Graphics Gems III
        b->fdir.x = b->dir.x - b->dir.y * dtheta;
        b->fdir.y = b->dir.y + b->fdir.x * dtheta;

        b->fperp     = V2f(-b->fdir.y, b->fdir.x);

        //
        // Future positions (absolute) of ends one and two
        //
        b->fend1pos  = b->fpos + b->fdir * b->lengthOverTwo;
        b->fend2pos  = b->fpos - b->fdir * b->lengthOverTwo;

        //
        // Future velocities (absolute) of ends one and two
        //
        b->fend1vel  = b->fvel + b->fperp * (b->fomega * b->lengthOverTwo);
        b->fend2vel  = b->fvel - b->fperp * (b->fomega * b->lengthOverTwo);
    }

    xfirst = xfirst ? false : true;
}


void
Sim::calcCorrectionForces(float dt)
{
    for (Joint::container::iterator i = joints.begin(); i != joints.end(); ++i)
    {
        Joint *j = *i;
        Body *b1 = bodies[j->b1];
        Body *b2 = bodies[j->b2];

        //
        // future offsets (body relative) of joint connections
        // 
        V2f foff1 = b1->fdir * j->b1off;
        V2f foff2 = b2->fdir * j->b2off;

        //
        // future positions (absolute) of joint connections
        // 
        V2f fpos1 = b1->fpos + foff1;
        V2f fpos2 = b2->fpos + foff2;

        //
        // future velocities (absolute) of joint connections
        // 
        V2f fvel1  = b1->fvel + b1->fperp * (b1->fomega * j->b1off);
        V2f fvel2  = b2->fvel + b2->fperp * (b2->fomega * j->b2off);

        //
        // Calculate correction force as a damped spring with zero rest
        // length connecting the ends of the joints.
        //
        V2f F = (fpos1 - fpos2) * joint_k + (fvel1 - fvel2) * joint_b;

        //
        // Apply the correction force
        //
        b1->F -= F;
        b2->F += F;

        //
        // Apply the torque caused by the joint correction force using
        // the current, not future offsets.
        //
        V2f off1 = b1->dir * j->b1off;
        V2f off2 = b2->dir * j->b2off;

        //
        // N = r cross F
        //
        b1->N  += off1 % (-F);
        b2->N  += off2 % F;


        //
        // Now do the joint angle targeting
        //
        float theta_diff = b1->ftheta - b2->ftheta;
        if (j->sign * (b1->dir ^ b2->dir) > 0.99)
        {
            _selfCollided = true;
        }
        j->prev_theta = theta_diff;

        theta_diff -= j->target_theta 
            + j->target_exp->evaluate(getCurrentTime());

        float dtarget = fabs(j->target_exp->devaluate(getCurrentTime()));
        if (dtarget > j->max_dtarget) j->max_dtarget = dtarget;

        //
        // Normalize the angle to [-Pi, Pi]
        //
        while (theta_diff > Math::Pif)
        {
            theta_diff -= Math::TwoPif;
        }
        while (theta_diff < -Math::Pif)
        {
            theta_diff += Math::TwoPif;
        }

        float omegadiff = b1->omega - b2->omega;

        //
        // Apply the torque as a damped angular spring.
        //
        // I should really go and calculate the force that caused this
        // torque as well.
        //
        float N = joint_ang_k * theta_diff + joint_ang_b * omegadiff; 

        b1->N -= N;
        b2->N += N;
    }

    for (Body::container::iterator i = bodies.begin(); i != bodies.end(); ++i)
    {
        Body *b = (*i);

        if (b->fend1pos.y < ground_level)
        {
            V2f F;
            
            //
            // Treat the ground as a damped spring
            //
            F.y = (ground_level - b->fend1pos.y) * ground_k 
                  - b->fend1vel.y * ground_b;

            // F.x = something clever involving ground_friction

            //
            // cheesy version of friction - make the ground perfectly sticky
            // this should get fixed
            //
            //b->vel.x = b->fend1vel.y * b->dir.y;
            //F.x = -ground_friction * b->fend1vel.x;
            F.x = -(b->fend1vel.x * b->M * 0.1 / timestep);

            //
            // Apply the corrective force...
            //
            b->F += F;

            //
            // ...and its resulting torque
            //
            b->N += (b->dir * b->lengthOverTwo) % F;
        }
        if (b->fend2pos.y < ground_level)
        {
            V2f F;
            
            //
            // Treat the ground as a damped spring
            //
            F.y = (ground_level - b->fend2pos.y) * ground_k 
                  - b->fend2vel.y * ground_b;

            // F.x = something clever involving ground_friction

            //
            // cheesy version of friction - make the ground perfectly sticky
            // this should get fixed
            //
            //b->vel.x = - b->fend2vel.y * b->dir.y;
            F.x = -(b->fend2vel.x * b->M * 0.1 / timestep);
            
            //
            // Apply the corrective force...
            //
            b->F += F;

            //
            // ...and its resulting torque
            //
            b->N += (-b->dir * b->lengthOverTwo) % F;
        }
    }
}

void
Sim::integrate(float dt)
{
    float min_height = 100000.0f;

    for (Body::container::iterator i = bodies.begin(); i != bodies.end(); ++i)
    {
        Body *b = *i;

        //
        // Integrate position, velocity, angle, angular velocity
        //
        b->pos += b->vel * dt;
        b->vel += b->F * (dt * b->oneOverM);

        float dtheta = b->omega * dt;
        b->theta += dtheta;

        //
        // Recalculate normalized direction
        //
        //b->dir = V2f((float) cos(b->theta), (float) sin(b->theta));
        // sinusoidal circle drawing courtesy of Akin and Simar, 
        // Graphics Gems III
        b->dir.x = b->dir.x - b->dir.y * dtheta;
        b->dir.y = b->dir.y + b->dir.x * dtheta;

        //
        // Normalize theta to [-Pi, Pi] to aid the angle targeting
        //
        while (b->theta > Math::TwoPif)
        {
            b->theta -= Math::TwoPif;
        }
        while (b->theta < -Math::TwoPif)
        {
            b->theta += Math::TwoPif;
        }
        b->omega += b->N * dt * b->oneOverI;

        //
        // Recompute end positions
        //
        b->end1pos = b->pos + b->dir * b->lengthOverTwo;
        b->end2pos = b->pos - b->dir * b->lengthOverTwo;


        //
        // Stuff for the genetic algorithm
        // 
        if (b->end1pos.y < min_height) min_height = b->end1pos.y;
        if (b->end2pos.y < min_height) min_height = b->end2pos.y;
    }

    for (Body::container::iterator i = bodies.begin(); i != bodies.end(); ++i)
    {
        Body *b1 = *i;

        V2f myEnd11 = b1->pos + b1->dir * 0.9f * b1->lengthOverTwo;
        V2f myEnd12 = b1->pos + b1->dir * 0.9f * b1->lengthOverTwo;

        for (Body::container::iterator j = i; j != bodies.end(); ++j)
        {
            Body *b2 = *j;

            V2f myEnd21 = b2->pos + b2->dir * 0.9f * b2->lengthOverTwo;
            V2f myEnd22 = b2->pos + b2->dir * 0.9f * b2->lengthOverTwo;

            if (((myEnd21 - myEnd11) % b1->dir) 
                * ((myEnd22 - myEnd11) % b1->dir) < 0.0f)
            {
                if (((myEnd11 - myEnd21) % b2->dir) 
                    * ((myEnd12 - myEnd21) % b2->dir) < 0.0f)
                {
                    _selfCollided = true;
                }
            }
        }
    }

    if (min_height > _maxHeight) _maxHeight = min_height;
}

void
Sim::doTimestep()
{
    calcConstantForces(timestep);
    predictFuturePositions(timestep);
    calcCorrectionForces(timestep);
    integrate(timestep);

    _currentTime += timestep;
}

