#include <GL/glut.h>
#include <GL/gl.h>

#include <sys/time.h>
#include <unistd.h>

#ifndef _Body_h_
# include "Body.h"
#endif

#ifndef _Joint_h_
# include "Joint.h"
#endif

#ifndef _Sim_h_
# include "Sim.h"
#endif

#ifndef _Mouse_h_
# include "Mouse.h"
#endif

#ifndef _Math_h_
# include "Math.h"
#endif

#ifndef _Exp_h_
# include "Exp.h"
#endif

#ifndef _Parser_h_
# include "Parser.h"
#endif

#ifndef _GeomGL_h_
# include "GeomGL.h"
#endif

static bool do_strobe = false;
static int strobe;
static int strobe_count = 0;
static int strobe_total;
static float loop_time;
static float prevtime = 0;
static long sec_orig;
static struct timezone tz;
static struct timeval  tv; 
static const bool every_frame = false;
static Sim sim;

void
idle()
{
    if (do_strobe && strobe_count > strobe_total) 
    {
        return;
    }

    if (sim.getCurrentTime() > loop_time)
    {
        sim.reset();
    }

    if (every_frame)
    {
        sim.doTimestep();
    }
    else
    {
        if (do_strobe)
        {
            for (int i = 0; i < strobe; ++i)
            {
                sim.doTimestep();
            }
            strobe_count += 1;
        }
        else
        {
            gettimeofday(&tv, &tz);
            float time = (tv.tv_sec - sec_orig) + (tv.tv_usec * 0.000001f);
            while (prevtime < time)
            {
                sim.doTimestep();
                prevtime += sim.getTimestep();
            }
        }
    }

    glutPostRedisplay();
}

void
display()
{
    static bool cleared = false;
    if (!cleared) glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    if (do_strobe) cleared = true;

    glBegin(GL_LINES);
    glColor3f(0.0f, 0.3f, 0.3f);
    glVertex2f(-600.0f, sim.getGroundLevel());
    glVertex2f(600.0f, sim.getGroundLevel());
    
    if (do_strobe)
    {
        float tmp = (float) strobe_count / (float) strobe_total * 0.7f + 0.3f;
        glColor3f(tmp, tmp, tmp);
    }
    else
    {
        glColor3f(1.0f, 1.0f, 1.0f);
    }

    for (Body::container::iterator i = sim.bodies.begin(); 
         i != sim.bodies.end(); 
         ++i)
    {
        const Body *b = (*i);

        glVertex(b->end1pos);
        glVertex(b->end2pos);
    }

    if (do_strobe)
    {
        float tmp = (float) strobe_count / (float) strobe_total * 0.7f + 0.3f;
        glColor3f(tmp, 0.0f, tmp);
    }
    else
    {
        glColor3f(1.0f, 0.0f, 1.0f);
    }

    if (mouseDown)
    {
        glVertex(mousePos);
        glVertex(sim.bodies[0]->pos);
    }

    for (Joint::container::iterator i = sim.joints.begin(); 
         i != sim.joints.end(); 
         ++i)
    {
        Joint *j = *i;
        Body *b1 = sim.bodies[j->b1];
        Body *b2 = sim.bodies[j->b2];

        V2f pos1 = b1->pos + b1->dir * j->b1off;
        V2f pos2 = b2->pos + b2->dir * j->b2off;

        glVertex(pos1);
        glVertex(pos2);
    }

    cout << flush;

    glEnd();
    glutSwapBuffers();
}

float screenZoom;
int screenWidth;
int screenHeight;
float yOff = 1.0f;

void
initGL()
{
    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(45.0f, 1.0f, 0.0f, 10000.0f);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    gluLookAt(0.0f, screenZoom * 0.5f - 2.0f, screenZoom,
              0.0f, screenZoom * 0.5f - 2.0f, 0.0f,
              0.0f, 1.0f, 0.0f);
}

void
reshape(int width, int height)
{
    screenWidth = width;
    screenHeight = height;
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(55.0f, (float) width / (float) height, 1.0f, 10000.0f);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    gluLookAt(0.0f, screenZoom * 0.5f - yOff, screenZoom,
              0.0f, screenZoom * 0.5f - yOff, 0.0f,
              0.0f, 1.0f, 0.0f);
    glViewport(0, 0, width, height);
}

void
calcPullPos(int x, int y)
{
    
    mousePos.x = ((float) x / (float) (screenWidth - 1)) * screenZoom 
        - 0.5f * screenZoom;

    float shminus = (float) (screenHeight - 1);
    mousePos.y = ((shminus - (float)y) / shminus) * screenZoom 
        - yOff;
}

void
mouse(int button, int state, int x, int y)
{
    if (state == GLUT_DOWN)
    {
        mouseDown = true;
        calcPullPos(x, y);
    }
    else
    {
        mouseDown = false;
    }
}

void
mouseMotion(int x, int y)
{
    calcPullPos(x, y);
}

void
keyboard(unsigned char key, int x, int y)
{
    switch (key)
    {
      case 'a':
        screenZoom *= 2.0f;
        reshape(screenWidth, screenHeight);
        break;
      case 's':
        screenZoom *= 0.5f;
        reshape(screenWidth, screenHeight);
        break;
    }
}

void
init()
{
    initGL();

    // we don't care about the actual timezone
    tz.tz_minuteswest = 0;
    tz.tz_dsttime = 0;

    gettimeofday(&tv, &tz);
    sec_orig = tv.tv_sec;
    prevtime = tv.tv_usec * 0.000001f;

    screenZoom = 10.0f;
    mouseDown = false;
}

int
main(int argc, char *argv[])
{

    if (argc < 2 || argc > 4)
    {
        cout << "Usage: " << argv[0] << " <simfile> [loop time] [strobe]" << endl;
        return -1;
    }

    int error = sim.parse(argv[1]);

    if (error != 0)
    {
        cerr << "<main> Parse error, exiting" << endl;
        return error;
    }
    
    if (argc >= 3)
    {
        loop_time = atof(argv[2]);
        cout << "Viewer: looping after " << loop_time << " seconds" << endl;
    }
    else
    {
        loop_time = 100000.0f;
    }

    if (argc >= 4)
    {
        do_strobe = true;
        float tmp_strobe = atof(argv[3]);
        strobe = (int) (tmp_strobe / sim.getTimestep());
        strobe_count = 0;
        strobe_total = (int) (loop_time / tmp_strobe);
        cout << "Viewer: strobing with interval " << tmp_strobe << endl;
    }
    else
    {
        do_strobe = false;
    }    

    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB);
    glutCreateWindow("luxo");
    glutDisplayFunc(display);
    glutKeyboardFunc(keyboard);
    glutMouseFunc(mouse);
    glutMotionFunc(mouseMotion);
    glutReshapeFunc(reshape);
    glutIdleFunc(idle);

    init();

    glutMainLoop();

    return 0;
}

